/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * This file contains main program and all functions to create and handle the
 * interface editor (main event loop)
 *****************************************************************************/

#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sgtty.h>

#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>

#include <exBasics.h>
#include <exMsg.h>
#include <exEdit.h>
#include <exLayout.h>
#include <exErrors.h>
#include <exStxt.h>
#include <exMenu.h>
#include "cmds.h"
#include <ife.h>
#include "ifedit.h"
#include "version.h"
#include <icons/ife>

static char SccsID[] = "@(#)ife.c	1.3\t1/27/93";            /* sccs keyword */

static struct IFEditor *ifedit;

static Display *display;
static Widget top;
static int dpyno;
static fd_set readfs[2];

static char fullName[256];
static char buf[512];
static char msg[100];
static void openEditor();

static struct IFEditor *createIFeditor();

static processMessage();
static forcedQuit();
static ifedTitle();
static processApplicationLine();
static closeIFEditor();
static startEditor();
static quitEditor();
static getApplByPid();
static void ifeInit();
static quitIFEditor();
static char* getApplication();

main(argc, argv)
     int argc;
     char *argv[];
{
  int maxfd = getdtablesize();
  int nfds, i, cnt;
  char *getenv();
  XEvent event;
  int eshErrorHandler();

  top = XtInitialize(NULL, "IFE", (XrmOptionDescRec *)NULL, 0, &argc, argv);

  display = XtDisplay(top);
  dpyno = ConnectionNumber(display);
  XSynchronize(display, True); 
  XSetErrorHandler(eshErrorHandler);

  signal(SIGHUP, forcedQuit);

  /* macro editor functions */
  cnt = 0;
  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0;
  cnt++;
  MH_FUNCTIONS[cnt].func = startEditor;
  MH_FUNCTIONS[cnt].funcName = "startEditor";
  cnt++;
  MH_FUNCTIONS[cnt].func = quitEditor;
  MH_FUNCTIONS[cnt].funcName = "quitEditor";
  cnt++;
  MH_FUNCTIONS[cnt].func = quitIFEditor;
  MH_FUNCTIONS[cnt].funcName = "quitIFEditor";
  cnt++;
  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0;


  ifeInit(display);
  ifedit = createIFeditor(top, 500, 500);
  initCmdTables(ifedit->notesEditor, CmdTable, EscCmds, ctrlXCmds, 
		(cmdDescr *)NULL);

  processInitFile(ifedit);
  openIFEditor(ifedit);
  
  set_wm_hints(ifedit);

  adjustReadMask(ifedit);

  setbuf(stdin, NULL);
  setbuf(stdout, NULL);
  
  while(True) {
    bcopy((char *) &readfs[0], (char *) &readfs[1], sizeof(readfs[0])); 
    XFlush(display);
    nfds = select(maxfd, &readfs[1], 
		  (fd_set *)0, (fd_set *)0, (struct timeval *)0); 
    if (nfds < 0) {
      exError("IFE", "select error !", NULL);
      continue;
    }
    if (FD_ISSET(dpyno, &readfs[1])) {			/* process events */
      while (XtPending()) {
	XtNextEvent(&event);
	XtDispatchEvent(&event);
      }
    } else if (FD_ISSET(fileno(stdin), &readfs[1])) 	/* from esh */
      processMessage(ifedit, fileno(stdin));
    else 				       	/* from one of the editors */
      for (i = 0; i < MDS_MAXAPPL; i++)
	if (ifedit->application[i].flags & IFE_STARTED &&
	    ifedit->application[i].fd != -1 &&
	    FD_ISSET(ifedit->application[i].fd, &readfs[1]))
	  if (processMessage(ifedit, ifedit->application[i].fd) < 0)
	    killEditor(ifedit, i);
  }
}

/* processMessage: ***********************************************************
 * process a message that has been sent either by the esh or by one of the 
 * editors. Messages are sent by the esh to inform the interface editor if a
 * component has been started or closed so that the local data-structures 
 * can be updated.
 *****************************************************************************/

static processMessage(ifed, fd)
     struct IFEditor *ifed;
     int fd;
{
  int pid, slot, rc, opc, info, cnt;
  extern errno;
  Arg args[10];
  XmFontList font_list;

  font_list = 
    XmFontListCreate(exEditorFont(display), XmSTRING_DEFAULT_CHARSET);

  ifeClearMsg();

  if (exReceiveMsg(fd, &pid, &opc, &info) < 0) {
    /*sprintf(msg, "pid=%d, opc=%d, info=%d, errno=%d", pid, opc, info, errno);
    exError("IFE", "illegal message received: ", msg, NULL);*/
    return(-1);
  }

  /*sprintf(msg, "pid=%d, opc=%d, info=%d", pid, opc, info);
  exError("IFE", "message received: ", msg, NULL);*/
  
  switch (opc) {
  case IFE_INFO:			/* info about components from esh*/
    switch (info) {
    case IFE_START:			/* component has been started */
      if ((slot = getFreeSlot(ifed)) < 0) {
	ifeMsg("No free slot !");
	return(-1);
      }
      ifed->application[slot].pid = pid;
      ifed->application[slot].flags = IFE_OCCUPIED | IFE_STARTED;
      /*printf(msg, "application --%d-- started !", pid);
      exError("IFE", msg, NULL);*/
      break;
    case IFE_STOP:				/* component has been ended */
      if (( slot = getApplByPid(ifed, pid)) < 0) {
	ifeMsg("Illegal pid in message !");
	return(-1);
      }

      if (ifed->application[slot].flags & IFE_EDITOR) 
	killEditor(ifed, slot);
      else {
	/*exError("IFE", "application -- ", 
		ifed->application[slot].name, " -- stopped !", NULL); */
	ifed->application[slot].pid = 0;
	ifed->application[slot].flags = 0;
	free(ifed->application[slot].name);
	ifed->application[slot].name = (char *) 0;
      }
      break;
    default:
      ifeMsg("Illegal info !");
      break;
    }
    break;
  case EXMSG_NAME:				/* receive a component name */
    if ((rc = read(fd, buf, info)) != info) {
      ifeMsg("Illegal string in message !");
      return(-1);
    }
    buf[rc] = EOS;
    
    if ((slot = getApplByPid(ifed, pid)) < 0) {
      ifeMsg("Illegal pid in messaged !");
      return(-1);
    }
    ifed->application[slot].name = strCopy(buf);
    break;
  case IFE_OPEN:				/* editor has been mapped */
    if ((slot = getApplByPid(ifed, pid)) < 0) {
      ifeMsg("Illegal pid in messaged !");
      return(-1);
    }
    ifed->application[slot].flags |= IFE_MAPPED;
    if (ifed->application[slot].button) {	/* remove button */
      XtDestroyWidget(ifed->application[slot].button);
      ifed->application[slot].button = (Widget) NULL;
      ifed->application[slot].y = -1;
    }
    break;
  case IFE_CLOSE:				/* editor has been iconised */
    if ((slot = getApplByPid(ifed, pid)) < 0) {
      ifeMsg("Illegal pid in message !");
      return(-1);
    }
    if (! ifed->application[slot].button) {
      cnt = 0;
      XtSetArg(args[cnt], XmNfontList, font_list); cnt++;
      XtSetArg(args[cnt], XmNlabelString, 
	       XmStringCreate(ifedit->application[slot].name, 
			      XmSTRING_DEFAULT_CHARSET)); cnt++; 

      ifedit->application[slot].button = 
	XtCreateManagedWidget("IFE button", xmPushButtonWidgetClass, 
			    ifed->button_win, args, cnt);
      XtAddCallback(ifedit->application[slot].button, XmNarmCallback, 
		    openEditor, &ifed->application[slot]);

    }
    ifed->application[slot].flags &= ~IFE_MAPPED;
    XFlush(display);
    break;
  default:
    ifeMsg("Illegal opcode !");
    break;
  }
  return(0);
}

/* openIFEditor: **************************************************************
 * display the interface editor on the screen
 *****************************************************************************/

openIFEditor(ifed)
     struct IFEditor *ifed;
{
  if (! (ifed->flags & IFE_MAPPED)) {
    ifed->flags |= IFE_MAPPED;
    XtPopup(ifed->ifed_shell, XtGrabNone);
  }
}

/* createIFeditor: ***********************************************************
 * create an instance of the interface editor at the given location 
 *****************************************************************************/

static struct IFEditor *createIFeditor(parent, x, y)
     Widget parent;
     int x,y;
{
  struct IFEditor *ifed;
  EditorDescr descr;
  static ifedTitle();
  int cnt;
  Arg args[10];
  BufInfo *bufinfoptr();
  
  descr.name = "IFE-notes";
  descr.inputFnct = (int (*)()) 0;
  descr.file = (char *) 0;
  descr.buffer = (char **) 0;
  descr.bufferLines = 1000;
  descr.lines = 16;
  descr.cols = 40;
  descr.flags = ED_MSGWINDOW | ED_SCROLLBAR | ED_HELP;
  

  ifed = (struct IFEditor *) Malloc(sizeof(struct IFEditor));

  /* create main widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNtitle, VERSION); cnt++;
  XtSetArg(args[cnt], XmNiconName, "ERGO Interface Editor"); cnt++;
  XtSetArg(args[cnt], XmNiconX, 1060); cnt++;
  XtSetArg(args[cnt], XmNiconY, 570); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  
  ifed->ifed_shell = XtCreatePopupShell(VERSION,   
			   applicationShellWidgetClass, parent, args, cnt);
  
  /* create top-level widget */

  cnt = 0;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;

  ifed->ifed_window = XtCreateManagedWidget("form", xmFormWidgetClass, 
			               ifed->ifed_shell, args, cnt);

  /* create the menubar and load menus */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  ifed->menubar = 
    exCreateMenuBar(ifed->ifed_window, args, cnt, (void (*)()) 0);
		
  if (makeFullName(getMenuDir(), IFEMENUS, fullName, sizeof(fullName)) < 0) 
    return((struct IFEditor *) 0);

  if (! initMenus(ifed->menubar, ifed->ifed_window, fullName, 
	    			(MenuDescr *) 0, (int *) 0)) {
    sprintf(msg, "Cannot load menus -- %s -- : interface editor aborted", 
	    							 fullName);
    exAlert(parent, 100, 100, msg);
    exit(1);
  }

   /* create the message area */
  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  ifed->msgwin = XtCreateManagedWidget("msg", xmDrawingAreaWidgetClass,
					ifed->ifed_window, args, cnt);
  descr.msg = ifed->msgwin;

  /* create the row column widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, ifed->menubar->menubar); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNbottomWidget, ifed->msgwin); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNpacking, XmPACK_TIGHT); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
  ifed->rowcol = XtCreateManagedWidget("rowcol", xmRowColumnWidgetClass,
			ifed->ifed_window, args, cnt);


  /* create the button area */
  cnt = 0;
  ifed->lform =  XtCreateManagedWidget("lform", xmFormWidgetClass, 
			               ifed->rowcol, args, cnt);

  /* create button area title */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("Button Area", XmSTRING_DEFAULT_CHARSET)); cnt++; 
  ifed->button_title =  XtCreateManagedWidget("title", xmLabelWidgetClass,
			ifed->lform, args, cnt);

  /* create button area */
  cnt = 0;
  XtSetArg(args[cnt], XmNresizeWidth, False); cnt++;
  XtSetArg(args[cnt], XmNpacking, XmPACK_COLUMN); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, ifed->button_title); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  ifed->button_win = XtCreateManagedWidget("buttons", 
		  xmRowColumnWidgetClass, ifed->lform, args, cnt);

  /* create the notes area */
  cnt = 0;
  ifed->rform =  XtCreateManagedWidget("rform", xmFormWidgetClass, 
			               ifed->rowcol, args, cnt);

  /* create notes editor title */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("Notes Editor", XmSTRING_DEFAULT_CHARSET)); cnt++; 
  ifed->note_title=  XtCreateManagedWidget("title", xmLabelWidgetClass,
			ifed->rform, args, cnt);
  
  /* create notes editor */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, ifed->note_title); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  ifed->notesEditor = 
    	makeEditor(ifed->rform, "notes", &descr, args, cnt);

  /* set height of message area */
  cnt = 0;
  XtSetArg(args[cnt], XmNheight, 
	3 * exFontHeight(bufinfoptr(ifed->notesEditor)->font) / 2); cnt++;
  XtSetValues(ifed->msgwin, args, cnt);

  ifed->flags = 0;
  return(ifed);
}

/* forcedQuit: ****************************************************************
 * this function is called upon receipt of a SIGHUP signal from the eshell in
 * order to terminate; 
 *****************************************************************************/

static forcedQuit()
{
  quitIFEditor((RequestParm *) NULL);
}

/* killEditor ****************************************************************
 *
 ****************************************************************************/

killEditor(ifed, slot)
     struct IFEditor *ifed;
     int slot;
{
  if ((ifed->application[slot].flags & IFE_EDITOR) &&
         (ifed->application[slot].flags & IFE_STARTED)) {
    kill(ifed->application[slot].pid, SIGHUP);
    ifed->application[slot].pid = -1;
    ifed->application[slot].fd  = -1;
    ifed->application[slot].y = -1;
    ifed->application[slot].flags = IFE_OCCUPIED | IFE_EDITOR;
    if (ifed->application[slot].button) {
      XtDestroyWidget(ifed->application[slot].button);
     ifed->application[slot].button = (Widget) 0;
    }
    sprintf(msg, "editor -- %s -- terminated by IFE !", 
	    			ifed->application[slot].name);
    /*exError("IFE", msg, NULL);*/
    ifeMsg(msg);
  }
}

/* getApplication: ************************************************************
 * creates a menus of requested applications and lets the user select one; 
 * if started is true, a menu of already started applications is generated; 
 * otherwise a menu of not yet started applications is built. If all is false, 
 * only editors are reported.
 *****************************************************************************/

static char *getApplication(ifed, started, editors, mapped)
     struct IFEditor *ifed;
     Boolean started, 		/* true if started applications */
          editors,		/* true if only editors */
          mapped;		/* true -> mapped, false -> don't care */
{
  int cnt, i;
  static struct {char *name; int indx;} applTable[MDS_MAXAPPL];
  MHmenu *mptr;
  MHitem *iptr;
  int x,y;
  char *applIndex;
  char cntStr[5];
  exMenu *applList;

  getMouse_xy(display, &x, &y);

  for (i = 0, cnt = 0; i < MDS_MAXAPPL; i++) {
    if (! (ifed->application[i].flags & IFE_OCCUPIED))
      continue;

    if (started && ifed->application[i].flags & IFE_STARTED) {
      if ((editors && ifed->application[i].flags & IFE_EDITOR) ||
	  (! editors && ! (ifed->application[i].flags & IFE_EDITOR))) {
	if (! mapped || (mapped && ifed->application[i].flags & IFE_MAPPED)) {
	  applTable[cnt].name  = ifed->application[i].name;
	  applTable[cnt].indx = i;
	  cnt++;
	}
      }
    } else if (!started && !(ifed->application[i].flags & IFE_STARTED)) {
      if ((editors && ifed->application[i].flags & IFE_EDITOR) ||
	      (! editors && ! (ifed->application[i].flags & IFE_EDITOR))) {
	if (! mapped || (mapped && ifed->application[i].flags & IFE_MAPPED)) {
	  applTable[cnt].name  = ifed->application[i].name;
	  applTable[cnt].indx = i;
	  cnt++;
	}
      }
    }
  }

  if (cnt == 0) {
    ifeMsg("No applications !");
    return(NULL);
  }

  setWorkApplication();
  
  if (! (mptr = createMenu())) {
    setError(ERR_CREATE);
    return(NULL);
  }
  
  insertMenu(mptr);
  
  setSelCnt(mptr, 1);
  setSelType(mptr, MDS_CNT_EQ);
  initMenu(mptr, MDS_L_POPUP, 
	         started ? "loaded editors" : "editors");
  setMenuPos(mptr, x, y, MDS_POS_ABS);
  
  mptr->m_hlp->hlpmsg = strCopy("Please select one application.");
  mptr->m_key = EOS;
  
  for (i = 0; i < cnt; i++) {				/* make the items */
    iptr = newItem(mptr, applTable[i].name, EOS);	/* create item */
    sprintf(cntStr, "%d", applTable[i].indx);
    iptr->i_return->retPattern = strCopy(cntStr);
  }
  
  if (! (applList = exMakeMenu(NULL, NULL, ifed->ifed_window, mptr))) {
    ifeMsg("Internal error: canot create menu !");
    return(0);
  }

  exCallMenu(applList);
  applIndex = strCopy(exGetMenuReturnValue(applList));
  freeMenu(applList);

  return(applIndex);
}

/* startEditorApplication: ****************************************************
 * fork the editor as a child of the esh; returns the pid of the child process
 *****************************************************************************/

startEditorApplication(ifed, slot, procName, parms)
     struct IFEditor *ifed;
     int slot;
     char *procName;		/* editor's name */
     char *parms[];
{
  struct ptydesc pt;
  struct sgttyb ttyp;
  int pid;

  ifed->application[slot].pid = -1;

  if (openpty(&pt) < 0)
    return(-1);

  /* set cbreak-mode for both ends of the pty */
  if (ioctl(pt.pt_pfd, TIOCGETP, &ttyp) < 0)
    return(-1);
  ttyp.sg_flags |= RAW;
  if (ioctl(pt.pt_pfd, TIOCSETP, &ttyp) < 0)
    return(-1);
  if (ioctl(pt.pt_tfd, TIOCGETP, &ttyp) < 0)
    return(-1);
  ttyp.sg_flags |= RAW;
  if (ioctl(pt.pt_tfd, TIOCSETP, &ttyp) < 0)
    return(-1);

  if ((pid = vfork()) < 0)    
    return(-1);

  if (pid == 0) {		/* the son */
      close(pt.pt_pfd);
      dup2(pt.pt_tfd, 0);		
      dup2(pt.pt_tfd, 1);
/*      dup2(pt.pt_tfd, 2);  */
      close(pt.pt_tfd);

      execvp(procName, parms);
      
      fatal("IFE", "cannot start editor -- ", procName, " --", NULL);
      exit(1);
    }

  ifed->application[slot].pid = pid;
  ifed->application[slot].fd = pt.pt_pfd;
  ifed->application[slot].flags |= IFE_STARTED;
  close(pt.pt_tfd);
  adjustReadMask(ifed);
  /*sprintf(msg, "pid = %d, fd=%d", pid, pt.pt_pfd);
  exError("IFE", "editor -- ", procName, 
	  " -- started by Interface Editor:", msg, NULL);*/
  return(pid);
}

/* getFreeSlot: **************************************************************
 * get a free slot in the listeners application table.
 *****************************************************************************/

getFreeSlot(ifed)
     struct IFEditor *ifed;
{
  register i;
  
  for (i = 0; i < MDS_MAXAPPL; i++) {
    if (! (ifed->application[i].flags & IFE_OCCUPIED)) {
      return(i);
    }
  }
  
  setError(ERR_MAX);
  return(-1);
}


/* adjustReadMask: ***********************************************************
 * adjust the mask used to listen to the file-descriptors of all connected 
 * applications.
 *****************************************************************************/

adjustReadMask(ifed)
     struct IFEditor *ifed;
{
  register i;

  FD_ZERO(&readfs[0]);
  FD_SET(dpyno, &readfs[0]);		/* connection with X-server */
  FD_SET(fileno(stdin), &readfs[0]);	/* stdin for connection with esh */

  for (i = 0; i < MDS_MAXAPPL; i++) {	/* look for connected editors */
    if (ifed->application[i].flags & IFE_OCCUPIED)
      FD_SET(ifed->application[i].fd, &readfs[0]);
  }
}

/* ifeMsg: ********************************************************************
 * forward a message to the user. If a msg-window has been defined, the 
 * message is displayed there, otherwise a notifier is displayed.
 *****************************************************************************/

ifeMsg(str)
     char *str;
{
  if (ifedit)
    editorMsg(ifedit->notesEditor, str);
  else
    exNotifyUser(top, -1, -1, str);
}


/* ifeClearMsg: **************************************************************
 * clear the message window if one has been defined
 ****************************************************************************/

ifeClearMsg()
{
  if (ifedit)
    editorClearMsg(ifedit->notesEditor);
}

/* processInitFile: **********************************************************
 * process each line of the user's init file '.iferc' if existing.
 ****************************************************************************/

int processInitFile(ifed)
     struct IFEditor *ifed;
{
  FILE *f;
  char *home;
  char *getenv();
  int slot;

  if (! (home = getenv("HOME"))) {
    setError(ERR_NULL);
    return(-1);
  }

  if (makeFullName(home, ".iferc", buf, sizeof(buf)) < 0)
    return(-1);

  if (! (f = fopen(buf, "r")))
    return(-1);

  while (fgets(buf, sizeof(buf), f)) {
    buf[strlen(buf) - 1] = EOS;
    if ((slot = getFreeSlot(ifed)) != -1) 
      processApplicationLine(ifed, slot, buf);
  }
  
  return(fclose(f));
}
    
/* processApplicationLine: ****************************************************
 * process one line of the init file  and start the corresponding 
 * editor.
 * Each line in the init file has the following syntax:
 * [flag] application [parameters]
 * where flag is none or one of the following characters:
 *     '+'     to indicate that the editor should be started automatically by 
 *             the interface editor
 *****************************************************************************/

static processApplicationLine(ifed, slot, line)
     struct IFEditor *ifed;
     int slot;
     char *line;
{
  char *parms[IFE_MAXPARMS];
  Boolean dostart = FALSE;
  register char *ptr;
  char *procName;
  register i, cnt;
  char *strtok(), *rindex();
  
  if (*line == '+') {			/* start application */
    line++;
    dostart = True;
  }
  
  /* parse the string containing the application name and eventually the 
     parameters for the call 
  */
  
  for (i = 0, ptr = strtok(line, " "); ptr && i < IFE_MAXPARMS; 
       						ptr = strtok(NULL, " "))
    parms[i++] = ptr;
  
  cnt = i;
  parms[cnt] = NULL;
  
  ifed->application[slot].fullName = strCopy(parms[0]);

  procName = strCopy(parms[0]);	      	/* full application name */

  if (ptr = rindex(parms[0], '/'))
    parms[0] = ++ptr;

  if (! ifed->application[slot].name)
    ifed->application[slot].name = strCopy(parms[0]);

  ifed->application[slot].flags |= (IFE_OCCUPIED | IFE_EDITOR);

  if (dostart)
    startEditorApplication(ifed, slot, procName, parms);
  
  free(procName);
  return(slot);
}

/******************************************************************************
 * 
 *                              IFE - menu 
 *
 *****************************************************************************/

/* quitIFEditor: **************************************************************
 * quit interface editor
 *****************************************************************************/
/*ARGSUSED*/
static quitIFEditor(parms)
     RequestParm *parms;
{
  register i;

  for (i = 0; i < MDS_MAXAPPL; i++)		/* kill all children */
    killEditor(ifedit, i);

  if (needSave(ifedit->notesEditor))
    editExit(ifedit->notesEditor);
 
  exit(1);
}

/******************************************************************************
 * 
 *                              editors - menu 
 *
 *****************************************************************************/

/* openEditor ****************************************************************
 * open (map) the corresponding editor 
 *****************************************************************************/

/*ARGSUSED*/
static void openEditor(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  struct ProcDescr *p = (struct ProcDescr *) client_data;

  if (p->flags & IFE_MAPPED)
    return;

  exSendMsg(p->fd, p->pid, IFE_OPEN, 0);
  p->flags |= IFE_MAPPED;
  if (p->button) {				/* remove button */
    XtDestroyWidget(p->button);
    p->button = (Widget) NULL;
    p->y = -1;
  }
}

/* startEditor: ***************************************************************
 * start an editor application 
 *****************************************************************************/
/*ARGSUSED*/
static startEditor(parms)
     RequestParm *parms;
{
  char *indx;
  int slot;
  char *parm[3];

  if (! (indx = getApplication(ifedit, False, True, False)))
    return;

  slot = atoi(indx);
  if (slot < 0 || slot > MDS_MAXAPPL)
    return;
  
  parm[0] = ifedit->application[slot].fullName;
  parm[1] = "-c";
  parm[2] = NULL;
  startEditorApplication(ifedit, slot, parm[0], parm);
}

/* quitEditor: ****************************************************************
 * quit the corresponding editor process 
 *****************************************************************************/
/*ARGSUSED*/
static quitEditor(parms)
     RequestParm *parms;
{
  char *indx;
  int slot;

  if (! (indx = getApplication(ifedit, True, True, False)))
    return;

  slot = atoi(indx);
  if (slot < 0 || slot > MDS_MAXAPPL)
    return;
  
  killEditor(ifedit, slot);
}

static getApplByPid(ifed, pid)
     struct IFEditor *ifed;
     int pid;
{
  register i;
  for (i = 0; i < MDS_MAXAPPL; i++) 
    if (ifed->application[i].pid == pid)
      return(i);

  return(-1);
}

/*ARGSUSED*/
int eshErrorHandler(dsp, event)
     Display *dsp;
     XErrorEvent *event;
{
  kill(getpid(), SIGQUIT);
}

/* ifeInit: ******************************************************************
 * initialize fonts and menu handling 
 ****************************************************************************/

static void ifeInit(d)
     Display *d;
{
  exInitLayout(d);
  exInitMenus(d);
}

/* set_wm_hints: ************************************************************
 * specify desired presentation for window manager
 ****************************************************************************/

set_wm_hints(ifed)
     struct IFEditor *ifed;
{
  XWMHints wm_hints;

  wm_hints.flags = IconPixmapHint | IconPositionHint;
  wm_hints.icon_pixmap = 
    XCreateBitmapFromData(XtDisplay(ifed->ifed_shell), 
			  XDefaultRootWindow(XtDisplay(ifed->ifed_shell)),
			  ife_bits, ife_width, ife_height);
  wm_hints.icon_x = 1060;
  wm_hints.icon_y = 570;

  XSetWMHints(XtDisplay(ifed->ifed_shell), 
	      XtWindow(ifed->ifed_shell), &wm_hints);
}
